#Complete SMC samplers algorithm for GenHMSAR data

SMC_complete_HMSAR_GenStates<-function(N=1000, B=100, temp.sched, data, prop.var.vec, hyper.vec, X.init.vec=c(1,0), rval, no.X, OriginalModel=TRUE, reduc=FALSE){
#SMC samplers for a Generalised Hamilton Markov Switching Autoregressive Model (GenHMSAR) where both the mean and variance are permitted to switch
    #N= number samples used to approximate the distributions in question
    #B = number of distributions to be approximated, i.e. number of iteration steps to move from prior to posterior
    #temp.sched = specified tempering schedule which controls how quickly we move between iterations /how dissimilar they are (0, 1)
        #Default is a simple linear schedule
	#data=a vector which represents the sequence of observations we are analysing
	#prop.var.vec= proposal variance vector for TP, means, precisions and AR 
    #hyper.vec= hyper parameters for transition probabilities, means and precisions
	#X.init.vec=probability inialisation vector for the underlying MC
    #rval = order of autoregressive nature on previous observations
	#no.X=the number of underlying states in the HMM
	#OriginalModel= TRUE (original HMSAR considered which only permits change in means), FALSE (changes in both mean and variance are permitted)
#reduc=TRUE (reduction is performed on the expanded SS which ommits tuple moves which are very unlikely) or FALSE (no reduction on expanded SS)

#Initialisation, obtain a starting set of particles/samples, log lik and their weights
SMC_Init_info<-SMC_intialisation_HMSAR_GenStates(N=N, data=data, hyper.vec=hyper.vec, init.vec=X.init.vec, rval=rval, no.X=no.X, OriginalModel=OriginalModel, reduc=reduc)

init.samples<-SMC_Init_info$I_Samples #Samples from the prior
init.loglik<-SMC_Init_info$I_Log; init.weights<-SMC_Init_info$I_Weights
	#The associated log lik and weights

init.samples.mat<-matrix(init.samples, nrow=N, ncol=1, byrow=F) #Matrix where each row represents a different sample
    #The idea is store all samples in an array where the 3rd dimension represents the iterations step

#Storage
particle.storage.array<-array(init.samples.mat, dim=c(N, 1, B)) # In array form so that the 3rd dimensions represents the iteration
loglik.mat<-t(as.matrix(init.loglik)) #Each row of log lik and weights mat will represent different iterations, each column represents a different sample
nweights.mat<-t(as.matrix(init.weights))
resample.vec<-NULL #Iterations at which we resample
NC.vec<-NULL #The normalising constants for the weights we are used to approximate the normalising constants of distributions

#Mutation and Reweighting Reiteration step
for(iter in 2:B){

temperature<-temp.sched[iter] #Current temp
curr.part.mat<-particle.storage.array[,,iter-1] #Current set of particles to mutate
curr.loglik.vec<-loglik.mat[iter-1,] #Their likelihoods
curr.nweights.vec<-nweights.mat[iter-1,] #Their normalised weights

#Resample?
ESS<-1/sum(curr.nweights.vec^2) #Effective Sample size to determine whether resampling is necessary
if({ESS<(N/2)} & {iter !=B}){
Resample_indices<-System_Resample(index.vec=1:N, w.vec=curr.nweights.vec) #Resampling indices particles, utilising systematic resampling
curr.part.mat<-curr.part.mat[Resample_indices] #Extract the appropriate "strong" particles
curr.loglik.vec<-curr.loglik.vec[Resample_indices] #...and their associated loglike
curr.nweights.vec<-1/N #All resampled samples are equally weighted
resample.vec<-c(resample.vec, iter) #Add the iteration time to the set of resample time points
}

#Weights are independent of the new mutated samples
inc.weights<-exp((temperature-temp.sched[iter-1])*curr.loglik.vec) #Incremental weights if we're not resampling
n.weights<-curr.nweights.vec*inc.weights #Unnormalised weights...
NC<-sum(n.weights)
	NC.vec<-c(NC.vec, NC) #Add normalising constant of weights to storage as they'll be useful in approximating normalising constants of distributions in the sequence
n.weights<-n.weights/NC #...now normalised
nweights.mat<-rbind(nweights.mat, n.weights) #Add to storage

#Actual mutation procedure
for(i in 1:N){
curr.part<-curr.part.mat[[i]]; curr.loglik<-curr.loglik.vec[i] #Single sample and loglik of interest
mutation_info<-SMC_mutation_HMSAR_GenStates(curr.theta=curr.part, curr.li=curr.loglik, prop.var.vec=prop.var.vec, hyper.vec=hyper.vec, temp=temperature, data=data, init.vec=X.init.vec, iter=iter, OriginalModel=OriginalModel, reduc=reduc)
#Mutation information of the current particle

curr.part.mat[[i]]<-mutation_info[[1]] #Add new mutated sample to particle matrix
curr.loglik.vec[i]<-mutation_info[[2]] #Similarly for the log likelihood
}

#Adding everything storage (samples, log lik) and reweighting accordingly
particle.storage.array[,,iter]<-curr.part.mat
loglik.mat<-rbind(loglik.mat, curr.loglik.vec)
}

list(curr.part.mat, n.weights, curr.loglik.vec, NC.vec, resample.vec)
#Output, curr.part.mat=the current set of samples which approximate the final distribution in the sequence (the posterior)
#n.weights=the associated weights of these samples
#curr.loglik.vec=the assoiated log-lik of these samples
#NC.vec=the vector of normalising constants of weights at iteration which are then used to approximate normalising constants of distributions.
#resample.vec= vector containg the resampling times (if any)
}
