#SMC_prior_sampling function

SMC_prior_sampling_HMSAR_GenStates<-function(N=1000, hyper.vec, rval, no.X=2, OriginalModel=TRUE){
#Prior sampling function of the SMC samplers for a GenHMSAR
#For general number of states

    #N= number samples used to approximate the distributions in question
    #hyper.vec= hyper parameters for transition probabilities, means and precisions
#rval = (Effective) Order of dependency introduced by AR parameter
  #Effective as in the case of rval=0 i.e. no dependency on previous observation, I might set rval=1 but the AR coefficients value is set to 1
#no.X=the number of underlying states in the HMM
	#OriginalModel= TRUE (original HMSAR considered which only permits change in means), FALSE (changes in both mean and variance are permitted)

hyper.info<-hyper.vec.sep(hyper.vec, no.X) #Function to seperate hyperparameters for each parameter
  
tp.hyper<-hyper.info[[1]] #Hyper parameters for transition probabilities
mu.hyper<-hyper.info[[2]] #Hyper parameters for emission means, mean and variance
prec.hyper<-hyper.info[[3]] #Hyper parameters for the emission precisions, the shape and scale 

#Transition Probability sampling
tp.hyper.mat<-matrix(1, nrow=no.X, ncol=no.X) # Matrix containing the corresponding hyper parameters for each transition probability vector
diag(tp.hyper.mat)<-tp.hyper #Setting the desired hyperparamerter strength

tp.init.array<-array(1, dim=c(no.X, N, no.X)) #An array to store the intialisation tranisition probability matrix information
for(row in 1:no.X){
tp.init.array[,,row]<-dirchlet.sample(shape.vec=tp.hyper.mat[row,] , no.X=no.X, no.samp=N) #Sample some row transition probabilities with respect to corresponding Dirichlet prior
}

#Means  sampling
mu.init.mat<-matrix(rnorm(N*no.X, mean=mu.hyper[1], sd=sqrt(mu.hyper[2])), nrow=no.X, ncol=N)
	#Matrix containing sampled mu's from prior. Each column denotes a different sample, each row is the associated mean for that particular state
#May want to edit so we can have different hyperparameters for the mean corresponding to each state

#Precision sampling
if(OriginalModel==TRUE){lambda.init.mat<-matrix(rgamma(N, shape=prec.hyper[1], scale=prec.hyper[2]), nrow=no.X, ncol=N, byrow=T)}
else{lambda.init.mat<-matrix(rgamma(N*no.X, shape=prec.hyper[1], scale=prec.hyper[2]), nrow=no.X, ncol=N)}
		#Matrix containing sampled lambda's from prior. Each column denotes a different sample, each row is the associated precision for that particular state.
			#For the original HMS-AR, we assume that there is a global precision which is not state dependent, and thus only sample one set of samples, but then replicate as necessary
	
#PAC sampling for AR coefficients, assume a uniform prior on (-1,1) for now. They are constrained to lie within the unit circle due to stationarity associated with AR parameter
  #This prior distribution on the AR coefficients can of course be changed dependent on ones belief
if(rval!=0){PAC.vec<-runif(rval*N, min=-1, max=1)} 
else {PAC.vec<-rep(0, N); rval<-rval+1}
PAC.mat<-matrix(PAC.vec, nrow=rval, ncol=N)

#Parameters in a list form where each element represents a different sample consisting of a transition probability matrix and precisions
X.SS<-1:no.X
theta.list<-vector("list", N)
for(part in 1: N){
tp.mat<-matrix(tp.init.array[,part,], nrow=no.X, ncol=no.X, byrow=T, dimnames=list(X.SS, X.SS)) 
	#Create a tranistion matrix from what we've sampled so far

theta<-list("TP matrix"=tp.mat, "Means"=mu.init.mat[,part], "Precisions"=lambda.init.mat[,part], "PAC"=PAC.mat[,part]) 
	#Form the appropriate theta by extracting the corresponding means, precision and PAC
theta.list[[part]]<-theta
}
theta.list #A list containing samples from the prior
}


############
#SMC initialisation
SMC_intialisation_HMSAR_GenStates<-function(N=1000, data, hyper.vec, init.vec, rval, no.X=2, OriginalModel=TRUE, reduc=FALSE){
#Performing the entire initialisation process of the SMC samplers
   #Prior sampling, likelihood calculation, weights

    #N= number samples used to approximate the distributions in question
	#data=a vector which represents the sequence of observations we are analysing
    #hyper.vec= hyper parameters for transition probabilities, means and precisions
	#init.vec=probability inialisation vector for the underlying MC
    #rval = order of autoregressive nature on previous observations
	#no.X=the number of underlying states in the HMM
	#OriginalModel= TRUE (original HMSAR considered which only permits change in means), FALSE (changes in both mean and variance are permitted)
#reduc=TRUE (reduction is performed on the expanded SS which ommits tuple moves which are very unlikely) or FALSE (no reduction on expanded SS)


N.star<-N
init.log.lik.vec<-NULL; init.theta.list<-NULL #Storage for log-like and theta that we are about to sample

while(N.star!=0){ #This whole procedure ensures the initial samples we begin with have a defined likelihood
Raw.particles<-SMC_prior_sampling_HMSAR_GenStates(N=N.star, hyper.vec=hyper.vec, rval=rval, no.X=no.X, OriginalModel=OriginalModel) #Perform some prior sampling

#Compute the log lik of each sample
Raw.log.lik<-sapply(Raw.particles, FUN=Cver.LogLik.HMSAR, data=data, raw.init.vec=init.vec, reduc=reduc)

bad.index<-which(is.nan(Raw.log.lik)==TRUE) #Which are bad samples with undefined log likelihood
if(length(bad.index)!=0){ #The case when bad samples exist
#Add good samples to appropriate storage units for keep
init.theta.list<-c(init.theta.list, Raw.particles[-bad.index])
init.log.lik.vec<-c(init.log.lik.vec, Raw.log.lik[-bad.index])
N.good<-N.star-length(bad.index) #How many good samples are there?
N.star<-N.star-N.good #New number of samples that we need obtain for initialisation
}
else{ init.theta.list<-c(init.theta.list, Raw.particles) #The case when there are no bad samples at all
init.log.lik.vec<-c(init.log.lik.vec, Raw.log.lik)
N.good<-N.star-length(bad.index) #How many good samples are there?
N.star<-N.star-N.good #New number of samples that we need obtain for initialisation}
}}
list("I_Samples"=init.theta.list, "I_Log Lik"=init.log.lik.vec, "I_Weights"=rep(1, N)/N)
#Output, list containg the samples from the prior, the associated log lik and their weights.
#Weights are normalised since we can sample from the prior directly.
}
