#Mutation functions for SMC samplers under GenHMSAR model

SMC_mutation_HMSAR_GenStates<-function(curr.theta, curr.li, prop.var.vec, hyper.vec, temp, data, init.vec, iter, OriginalModel=TRUE, reduc=FALSE){
#Function to perfrom on entire theta sample under a GenHMSAR model


#curr.theta=the current sample that we are going to mutate
#curr.li=the current log-lik associated with the sample we are about to mutate
    #data=a vector which represents the sequence of observations we are analysing
	#prop.var.vec= proposal variance vector for TP, means, precisions and AR 
    #hyper.vec= hyper parameters for transition probabilities, means and precisions
	#init.vec=probability inialisation vector for the underlying MC
	#OriginalModel= TRUE (original HMSAR considered which only permits change in means), FALSE (changes in both mean and variance are permitted)
#reduc=TRUE (reduction is performed on the expanded SS which ommits tuple moves which are very unlikely) or FALSE (no reduction on expanded SS)


#Proposal variance separation
tp.prop.var<-prop.var.vec[1]	#For the TP...
mu.prop.var<-prop.var.vec[2]	#...the emission means
prec.prop.var<-prop.var.vec[3]	#....the emission precisions
PAC.prop.var<-prop.var.vec[4]	#...the PAC

#Transition Probability 
curr.tp.mat<-curr.theta[[1]]; no.X<-nrow(curr.tp.mat)	#Current TP matrix and the number of states assumed in our underlying HMM

prop.tp.mat<-TPmat.proposal.logit(curr.TPmat=curr.tp.mat, prop.var=tp.prop.var)
#A proposal TP matrix based on the current TP matrix via a RWMH on a logit scale.

#Hyper parameters
hyper.info<-hyper.vec.sep(hyper.vec, no.X=no.X) #Hyperparameter separation
tp.hyper<-hyper.info[[1]] 
	tp.hyper.mat<-matrix(1, nrow=no.X, ncol=no.X); diag(tp.hyper.mat)<-tp.hyper #...TP in matrix form
mu.hyper.vec<-hyper.info[[2]]	#....for emission mean.
lambda.hyper.vec<-hyper.info[[3]]	#...for emission precision


#Setting up the particle so that we accept and reject each row of TP matrix one at a time
upd.theta<-curr.theta
upd.li<-curr.li

for(tp.row.index in 1:no.X){ #Test each TP row one by one
temp.prop.tp.mat<-upd.theta[[1]] #Extract the most recent version of the Transition Probability Vector
temp.prop.tp.mat[tp.row.index,]<-prop.tp.mat[tp.row.index,] #Add the suggested proposal TP row vector

acc_rej.info<-SMC_jointTP.AccRej_HMSAR_GenStates(curr.theta=upd.theta, curr.li=upd.li, prop.tp.mat=temp.prop.tp.mat, row=tp.row.index, phi=temp, data=data, init.vec=init.vec, tp.hyper.prior=tp.hyper.mat[tp.row.index,], reduc=reduc)
#Accept or refect it

upd.theta<-acc_rej.info[[1]] #The outcome sample
upd.li<-acc_rej.info[[2]] #The outcome associate log lik
}

#Each mean, proposal at the same time (independently), accept and reject separately
curr.mean.vec<-curr.theta[[2]] #Extract the current mean vector
prop.mean.vec<-curr.mean.vec + rnorm(no.X, mean=0, sd=sqrt(mu.prop.var)/(iter-1))  #Proposals for the means based on a RWMH

for(mean.index in 1:no.X){ #Testing each mean separately
inter.curr.mean.vec<-upd.theta[[2]]
inter.prop.mean.vec<-inter.curr.mean.vec
    inter.prop.mean.vec[mean.index]<-prop.mean.vec[mean.index] #Add the proposal mean in the corresponding place
    
acc_rej.info<-SMC_mean.AccRej_HMSAR_GenStates(curr.theta=upd.theta, curr.li=upd.li, prop.mean.vec=inter.prop.mean.vec, phi=temp, data=data, init.vec=init.vec, mean.prior.hyper=mu.hyper.vec, reduc=reduc)
#Perform the acceptance rejection mechanism for the means

upd.theta<-acc_rej.info[[1]] #The outcome sample
upd.li<-acc_rej.info[[2]]#... and it's associated log-likelihood
}

#Each precision, proposal at the same time (independently), accept and reject separately
curr.prec.vec<-curr.theta[[3]]
if(OriginalModel==TRUE){prop.prec.vec<-exp(log(curr.prec.vec)  +rnorm(1, mean=0, sd= sqrt(prec.prop.var/(iter-1))))
#Case when the precision is not actually state dependent, one proposal and acc/rej in general

#Proposal based on RWMH on log-scale centred around current precision value
acc_rej.info<-SMC_prec.AccRej_HMSAR_GenStates(curr.theta=upd.theta, curr.li=upd.li, prop.prec.vec=prop.prec.vec, phi=temp, data=data, init.vec=init.vec, prec.prior.hyper=lambda.hyper.vec, reduc=reduc)
   #Perform the acceptance rejection procedure

upd.theta<-acc_rej.info[[1]]#Acceptance sample
upd.li<-acc_rej.info[[2]]#....and it's associated log-lik
} 
else{prop.prec.vec<-exp(log(curr.prec.vec)  +rnorm(no.X, mean=0, sd= sqrt(prec.prop.var/(iter-1)))) #For the case when the precision is state dependent
	#Proposal is RWMH on log scale around current precision values

for(prec.index in 1:no.X){ #Testing each precision separately
inter.curr.prec.vec<-upd.theta[[3]] #Intermediary current precision, this takes into account the most recent values
inter.prop.prec.vec<-inter.curr.prec.vec
    inter.prop.prec.vec[prec.index]<-prop.prec.vec[prec.index] #Add the proposal precision in the corresponding place
    
acc_rej.info<-SMC_prec.AccRej_HMSAR_GenStates(curr.theta=upd.theta, curr.li=upd.li, prop.prec.vec=inter.prop.prec.vec, phi=temp, data=data, init.vec=init.vec, prec.prior.hyper=lambda.hyper.vec, reduc=reduc) 
   #Perform the acceptance rejection procedure

upd.theta<-acc_rej.info[[1]]#Acceptance sample
upd.li<-acc_rej.info[[2]]#...acceptance log-lik
}}

#Each PACF, proposal at the same time (independently), accept and reject separately
curr.PAC.vec<-curr.theta[[4]]; rval<-length(curr.PAC.vec)
if(curr.PAC.vec[1]==0){list("Mutated sample"=upd.theta, "Mutated log-lik"= upd.li)} #For the case when we have a 0th order HMS-AR effectively and we don't have to perform any proposal or testing. Function finishes here
else{prop.PAC.vec<-SMC_PAC.proposal(curr.PAC.vec=curr.PAC.vec, prop.PAC.var=PAC.prop.var/(iter-1))
	#When we have a proper HMS-AR and need to perform a proposal
	#This is proposal is based on a RWMH on a transformed logit scale
for(PAC.index in 1:rval){ #Test each PAC separately
inter.curr.PAC.vec<-upd.theta[[4]] #Intemediary current value of PAC vec, taking into account the most recent values
inter.prop.PAC.vec<-inter.curr.PAC.vec
  inter.prop.PAC.vec[PAC.index]<-prop.PAC.vec[PAC.index] #Add the proposal PAC to the corresponding place

acc_rej.info<-SMC_PAC.AccRej_HMSAR_GenStates(curr.theta=upd.theta, curr.li=upd.li, prop.PAC.vec=inter.prop.PAC.vec, phi=temp, data=data, init.vec=init.vec, reduc=reduc) #Perform the acceptance rejection procedure

upd.theta<-acc_rej.info[[1]] #Accepted sample
upd.li<-acc_rej.info[[2]]	#Accepted log-lik
}
list("Mutated sample"=upd.theta, "Mutated log-lik"= upd.li)
}

#Output, list containing the sample which has been mutated accordingly, and it's associated log-lik
}

#####################
SMC_jointTP.AccRej_HMSAR_GenStates<-function(curr.theta, curr.li, prop.tp.mat, row,phi, data, init.vec, tp.hyper.prior, reduc=FALSE){
#Acceptance rejection procedure with respect to the joint TP proposals
#curr.theta=current sample to mutate
#curr.li=log=lik associated with the sample
#prop.tp.mat=proposal TP mat to consider wheter to accept or reject
#row=the row of the TP matrix which is specically being considered
#phi =current temperature we are at
#data=time series analysed
#init.vec=probability initialisation vector for underlying MC
#tp.hyper.prior=hyper parameters associated with TP
#reduc= reduction on SS?

#Construct the proposal theta
prop.theta<-curr.theta
prop.theta[[1]]<-prop.tp.mat #Proposal theta with proposal TP mat

curr.tp.mat<-curr.theta[[1]]

#TP vectors that we are testing on
curr.tp.vec<-curr.tp.mat[row,]	#Current TP vector
prop.tp.vec<-prop.tp.mat[row,]	#Proposal TP vector

#Prior component
prop.prior<-dirchlet.log.density(xvec=prop.tp.vec, shape.vec=tp.hyper.prior) #Prior density of the proposal TP vector
if(is.nan(prop.prior)==TRUE){list(curr.theta, curr.li)} #Debugging one, reject if proposal prior is undefined
else{curr.prior<-dirchlet.log.density(xvec=curr.tp.vec, shape.vec=tp.hyper.prior) #Prior density of the current TP vector
prior.comp<-prop.prior-curr.prior #Prior component of acceptance probability

#Likelihood component
prop.li<-Cver.LogLik.HMSAR(data=data, theta=prop.theta, raw.init.vec=init.vec, reduc=reduc) #Log-lik under the proposal theta
if(is.nan(prop.li)==TRUE){list(curr.theta, curr.li)} #Debugging 2, reject if proposal lik is undefined
else{loglik.comp<-phi*(prop.li-curr.li) #Log-lik component of the acceptance probability

#Additional Jacobian component due to proposal on a transformation
prop.jacob<--sum(log(prop.tp.vec)) #Moving to the proposal TP FROM the current
curr.jacob<--sum(log(curr.tp.vec)) #Moving to the current TP FROM the proposal
jacob.comp<-curr.jacob-prop.jacob #Change of variables part of acceptance probability

alpha<-min(1, prior.comp+loglik.comp+jacob.comp) #Acceptance probability, working on a log scale
uno<-log(runif(1))
if(uno<alpha){list(prop.theta, prop.li)} #Accept proposal
else{list(curr.theta, curr.li)} #Reject proposal and take through the existing particle
}}
}



#########################
SMC_mean.AccRej_HMSAR_GenStates<-function(curr.theta, curr.li, prop.mean.vec, phi, data, init.vec, mean.prior.hyper, reduc=FALSE){
#Acceptance rejection procedure with respect to the means
#curr.theta=current sample to mutate
#curr.li=log=lik associated with the sample
#prop.mean.vec=proposal mean vec to consider wheter to accept or reject
#row=the row of the TP matrix which is specically being considered
#phi =current temperature we are at
#data=time series analysed
#init.vec=probability initialisation vector for underlying MC
#mean.prior.hyper=hyper parameters associated with mean
#reduc= reduction on SS?


#Construct the proposal theta
prop.theta<-curr.theta
prop.theta[[2]]<-prop.mean.vec #Create the proposal theta based on the proposal mean

curr.mean.vec<-curr.theta[[2]] #Exract the current mean vector

#Prior component
prop.prior<-sum(dnorm(prop.mean.vec, mean=mean.prior.hyper[1], sd=sqrt(mean.prior.hyper[2]), log=T)) #Prior density of the proposal means
if(is.nan(prop.prior)==TRUE){list(curr.theta, curr.li)} #Debugging one, reject if proposal prior is undefined
else{curr.prior<-sum(dnorm(curr.mean.vec, mean=mean.prior.hyper[1], sd=sqrt(mean.prior.hyper[2]), log=T))
prior.comp<-prop.prior-curr.prior #Prior component of acceptance probability

#Likelihood component
prop.li<-Cver.LogLik.HMSAR(data=data, theta=prop.theta, raw.init.vec=init.vec, reduc=reduc) #Log-lik of proposal sample
if(is.nan(prop.li)==TRUE){list(curr.theta, curr.li)} #Debugging 2, reject if proposal lik is undefined
else{loglik.comp<-phi*(prop.li-curr.li)
#Log-lik component of acceptance probability

#No additional change of variable component since we worked on the usual scale and no transformation took place

alpha<-min(1, prior.comp+loglik.comp) #Acceptance probability, working on a log scale
uno<-log(runif(1))
if(uno<alpha){list(prop.theta, prop.li)} #Accept proposal sample
else{list(curr.theta, curr.li)} #Reject proposal and take through the existing particle
}}
}


#########################
SMC_prec.AccRej_HMSAR_GenStates<-function(curr.theta, curr.li, prop.prec.vec, phi, data, init.vec, prec.prior.hyper, reduc=FALSE){
#Acceptance rejection procedure with respect to the precisions

#curr.theta=current sample to mutate
#curr.li=log=lik associated with the sample
#prop.prec.vec=proposal precision vector to consider whether to accept or reject
#phi =current temperature we are at
#data=time series analysed
#init.vec=probability initialisation vector for underlying MC
#prec.prior.hyper=hyper parameters associated with precision
#reduc= reduction on SS?


#Construct the proposal theta
prop.theta<-curr.theta
prop.theta[[3]]<-prop.prec.vec#Proposal sample with proposal precision in place

curr.prec.vec<-curr.theta[[3]]

#Prior component
prop.prior<-sum(dgamma(prop.prec.vec, shape=prec.prior.hyper[1], scale=prec.prior.hyper[2], log=T))
#Prior density of the proposal precision
if(is.nan(prop.prior)==TRUE){list(curr.theta, curr.li)} #Debugging one, reject if proposal prior is undefined
else{curr.prior<-sum(dgamma(curr.prec.vec, shape=prec.prior.hyper[1], scale=prec.prior.hyper[2], log=T)) #Prior density of current precision
prior.comp<-prop.prior-curr.prior #Prior component of acceptance ratio

#Likelihood component
prop.li<-Cver.LogLik.HMSAR(data=data, theta=prop.theta, raw.init.vec=init.vec, reduc=reduc) #Log-lik of proposal sample
if(is.nan(prop.li)==TRUE){list(curr.theta, curr.li)} #Debugging 2, reject if proposal lik is undefined
else{loglik.comp<-phi*(prop.li-curr.li) #Log-lik component of acceptance probability

#Additional Jacobian component due to proposal on a transformation, log scale
prop.jacob<-sum(log(1/prop.prec.vec))    #From current, jumping to proposal
curr.jacob<-sum(log(1/curr.prec.vec))    #From proposal, jumping to current
jacob.comp<-curr.jacob-prop.jacob #Change of variables component of acceptance probability

alpha<-min(1, prior.comp+loglik.comp+jacob.comp) #Acceptance probability, working on a log scale
uno<-log(runif(1))
if(uno<alpha){list(prop.theta, prop.li)} #Accept proposal
else{list(curr.theta, curr.li)} #Reject proposal and take through the existing particle
}}
}


#################
SMC_PAC.AccRej_HMSAR_GenStates<-function(curr.theta, curr.li, prop.PAC.vec, phi, data, init.vec, reduc=FALSE){
#Acceptance rejection procedure with respect to the PAC
#phi =current temperature we are at

#Construct the proposal theta
prop.theta<-curr.theta
prop.theta[[4]]<-prop.PAC.vec #Proposal sample with proposl PAC in place

curr.PAC.vec<-curr.theta[[4]]

#Prior component, since are dealing with uniform prior, the prior densities will be equal for both current and proposal location and cancle out
  #This of course changes if a more interesting prior is used.
prior.comp<-0

#Likelihood component
prop.li<-Cver.LogLik.HMSAR(data=data, theta=prop.theta, raw.init.vec=init.vec, reduc=reduc) #Log-lik under the proposal sample
if(is.nan(prop.li)==TRUE){list(curr.theta, curr.li)} #Debugging 2, reject if proposal lik is undefined
else{loglik.comp<-phi*(prop.li-curr.li) #Log-lik component of acceptance probability

#Additional Jacobian component due to proposal on a transformation, transformed logit
prop.jacob<-sum(log(2/((prop.PAC.vec+1)*(1-prop.PAC.vec))))   #At curr, jumping to prop
curr.jacob<-sum(log(2/((curr.PAC.vec+1)*(1-curr.PAC.vec))))   #At prop, jumping to curr
jacob.comp<-curr.jacob-prop.jacob #Jacobian component of acceptance probability

alpha<-min(1, prior.comp+loglik.comp+jacob.comp) #Acceptance probability, working on a log scale
uno<-log(runif(1))
if(uno<alpha){list(prop.theta, prop.li)} #Accept proposal
else{list(curr.theta, curr.li)} #Reject proposal and take through the existing particle
}
}
